// ===================================================================
//
// BBDrive_Test_5a.c
// Controller for Braitenberg Vehicle type 3 with intermediate neurons
// using shuntineg interation
// Based on Arduino Code keeping as much of this
// code as possible. Emulates ORP-12 and Parallax Servos
// Added GPS to log trajectory
// ==================================================================

#include <webots/robot.h>
#include <webots/motor.h> 
#include <stdio.h>
#include <webots/light_sensor.h>
#include <webots/gps.h>

#include "CBP_Helper_4.h"
#include "CBP_Plotter_5.h"

#define TIME_STEP 32
#define WHEEL_RADIUS 0.0205
#define AXLE_LENGTH 0.053
#define MAX_SPEED 6.28
#define PI 3.14159265358979

WbDeviceTag left_motor, right_motor;
WbDeviceTag ls0, ls7;
WbDeviceTag gps;

// Neural stuff -------------------------
float tau = 0.01;
float inL1, inR1;
float outL1, outR1;
float outL2, outR2;
float outIL, outIR;

float A = 250.0;
float B = 100.0;
float offset = 500;
float C = 25;
float D = 7.5;
float tonic = 0.25;
float E = 0.5;
float F = 0.15;

// ODE stuff ----------------
float u[6];
float dudt[6];
float deltaT = 0.01;

// Motor stuff --------------
float driveL, driveR;
float maxDrive = 60;

void deviceInit() {
 // get a handler to the motors and set target position to infinity (speed control).
  left_motor = wb_robot_get_device("left wheel motor");
  right_motor = wb_robot_get_device("right wheel motor");
  wb_motor_set_position(left_motor, INFINITY);
  wb_motor_set_position(right_motor, INFINITY);
  wb_motor_set_velocity(left_motor, 0.0);
  wb_motor_set_velocity(right_motor, 0.0);

  ls0 = wb_robot_get_device("ls0");
  wb_light_sensor_enable(ls0, TIME_STEP); 
  ls7 = wb_robot_get_device("ls7");
  wb_light_sensor_enable(ls7, TIME_STEP); 
  
  gps = wb_robot_get_device("gps");
  wb_gps_enable(gps, TIME_STEP); 
       
  }  

void compute_all_neurons(float inL, float inR){
  // Retinal layer
  dudt[0] = (-u[0] + inL - offset)/tau;
  dudt[1] = (-u[1] + inR - offset)/tau;
  u[0] += dudt[0] * deltaT;
  u[1] += dudt[1] * deltaT;
  outL1 = 1.0 / (1 + exp(-(u[0] - A)/B));
  outR1 = 1.0 / (1 + exp(-(u[1] - A)/B)); 

  // Intermediate layer - shunting inhibition
  dudt[2] = (-u[2] + tonic/(tonic + outL1))/tau;
  dudt[3] = (-u[3] + tonic/(tonic + outR1))/tau;
  
  //dudt[2] = (-u[2] + 1.0/(1.0 + 10*outL1))/tau;
  //dudt[3] = (-u[3] + 1.0/(1.0 + 10*outR1))/tau;
  u[2] += dudt[2] * deltaT;
  u[3] += dudt[3] * deltaT;
  
  outIL = 1.0 / (1 + exp(-(u[2] - E)/F));
  outIR = 1.0 / (1 + exp(-(u[3] - E)/F)); 

  // Motor neurons using shunting
  dudt[4] = (-u[4] + maxDrive*outIL)/tau;
  dudt[5] = (-u[5] + maxDrive*outIR)/tau;
  u[4] += dudt[4] * deltaT;
  u[5] += dudt[5] * deltaT;

  outL2 = maxDrive / (1 + exp(-(u[4] - C)/D));
  outR2 = maxDrive / (1 + exp(-(u[5] - C)/D)); 
  
}

int main(int argc, char **argv) {

  double time = 0;
  double omegaL, omegaR;
  double locX, locY;
  
  FILE* plotFile;
  plotFile = fopen("BBDrive_Test_5a.m", "w");
  writeOctaveHead(plotFile,A,B,C,D,E,F,offset,tonic); 
  
  wb_robot_init();
 
  deviceInit();

  while (wb_robot_step(TIME_STEP) != -1) {
          
    double LDR_L = wb_light_sensor_get_value(ls7);    
    double LDR_R = wb_light_sensor_get_value(ls0);    
    
    const double *gpsVals = wb_gps_get_values(gps); 
    locX = gpsVals[0];
    locY = gpsVals[1];
    
    // Emulate Arduino analogRead using ORP-12 LDRs
    inL1 = analogRead(LDR_L);
    inR1 = analogRead(LDR_R);

    compute_all_neurons(inL1, inR1);
   
    // Following for Vehicle 3a
    driveL = outL2;
    driveR = outR2;
    //
    // Following for Vehicle 3b
    //driveL = outR2;
    //driveR = outL2;
     
    // emulate Parallax Servos ----------------------
    // emulating deadband
    if(driveL < 10) driveL = 0;
    if(driveR < 10) driveR = 0;
    
    omegaL = (MAX_SPEED/50.0)*driveL;
    omegaR = (MAX_SPEED/50.0)*driveR;  
    
    // Useful alt prints to check behaviour of processing steps
    // printf("inL1 = %f,  inR1 = %f\n",inL1,inR1);
    // printf("avL = %f, avR = %f\n",avL,avR);   
    // printf("inL1 = %f, inR1 = %f, u[0] = %f, u[1] = %f\n",inL1,inR1,u[0],u[1]);
    // printf("outL1 = %f, outR1 = %f\n",outL1,outR1);
    // printf("u[2] = %f, u[3] = %f\n",u[2],u[3]);
    // printf("inL1 = %f, inR1 = %f, u[0] = %f, u[1] = %f\n",inL1,inR1,u[0],u[1]);
    //printf("inL1 = %f, inR1 = %f, outIL = %f, outIR = %f, outL2 = %f, outR2 = %f\n",
    //        inL1,inR1,outIR, outIL, outL2,outR2);
    printf("inL1 = %f, inR1 = %f, outL1 = %f, outR1 = %f, outIL = %f, outIR = %f, outL2 = %f, outR2 = %f\n",
            inL1,inR1, outL1, outR1,outIR, outIL, outL2,outR2);            
     //printf("outL2 = %f, outR2 = %f, drivvL = %f, driveR = %f, omegaL = %f, omegaR = %f\n",
    //        outL2,outR2,driveL,driveR,omegaL,omegaR);
    
    time += (double)(TIME_STEP);
        
    writeDataToOctave(plotFile,time, inL1,inR1, outL1,outR1, u[2],u[3],outIL, outIR, outL2,outR2,locX,locY);
          
     
    wb_motor_set_velocity(left_motor,  omegaL);
    wb_motor_set_velocity(right_motor, omegaR);

  };

  writeOctaveTail(plotFile);
  fclose(plotFile);
  wb_robot_cleanup();

  return 0;
}
