// ===================================================================
// File:          CBP_ePuck_Odometry_30
// Date:          01-09-23
// Status:        Devel
//
// Rotates on an arc of given radius for given angle using odometry
// then check accuracy of results with gps
// Uses simulated encoders, allows correction
// Based on CBP2403_D_ArcTest_1, with the Octave correction calcs
// Incorporated into this code.
// ==================================================================

#include <webots/robot.h>
#include <webots/motor.h> 
#include <webots/distance_sensor.h>
#include <webots/position_sensor.h>
#include <stdio.h>
#include <webots/gps.h>
#include <webots/compass.h>

#include "CBP_Plotter_2.h"
#include "CBP_Helper_2.h"

#define TIME_STEP 32
#define WHEEL_RADIUS 0.0205
#define AXLE_LENGTH 0.053
#define MAX_SPEED 6.28
#define PI 3.14159265358979

#define STATE_DRIVE 1
#define STATE_DONE 2

#define FLOAT_TO_INT(x) ((x)>=0?(int)((x)+0.5):(int)((x)-0.5))

// Global variables ----------------------------------------------

WbDeviceTag left_motor, right_motor;
WbDeviceTag comp;
WbDeviceTag gps;
WbDeviceTag leftWheelAngle, rightWheelAngle;
double thetaL, thetaR, dTheta;

void init() {

  left_motor = wb_robot_get_device("left wheel motor");
  right_motor = wb_robot_get_device("right wheel motor");
  wb_motor_set_position(left_motor, INFINITY);
  wb_motor_set_position(right_motor, INFINITY);
  wb_motor_set_velocity(left_motor, 0.0);
  wb_motor_set_velocity(right_motor, 0.0);

  gps = wb_robot_get_device("gps");
  wb_gps_enable(gps, TIME_STEP); 
  
  comp = wb_robot_get_device("comp1");
  wb_compass_enable(comp, TIME_STEP);    
  
    // get the wheel encoders
  leftWheelAngle = wb_robot_get_device("left wheel sensor");
  wb_position_sensor_enable(leftWheelAngle,TIME_STEP);

  rightWheelAngle = wb_robot_get_device("right wheel sensor");
  wb_position_sensor_enable(rightWheelAngle,TIME_STEP);
  
}

// ----------------------------------------------------------------
unsigned int ISR_Left() {

  unsigned int count;
  
  count = FLOAT_TO_INT((thetaL/dTheta));
  
  return count;
}

// ----------------------------------------------------------------
unsigned int ISR_Right() {

  unsigned int count;
  
  count = FLOAT_TO_INT((thetaR/dTheta));
  
  return count;
}

// ----------------------------------------------------------------

int main(int argc, char **argv) {

  double omegaL, omegaR;
  int state;
  

  double desRadius;
  double desDegrees;
  double desTheta;
  double desTime = 20.0;
  
  unsigned int nL, nR;
  unsigned int countL = 0;;
  unsigned int countR = 0;
  double nLf, nRf;
  double dsL, dsR, dsC;
  double thetaRob;
  
  double robotOmega;

  double percX, percY;
  
  double dx;
  
  FILE* plotFile;
  plotFile = fopen("Robot.m", "w");
  writeOctaveHead(plotFile);  
  
  dx = (2*PI*WHEEL_RADIUS)/1000;
  dTheta = 2*PI/1000;
  
  // Describe the desired arc
  desRadius = 0.300;  // metres
  desDegrees = 90.0;
  
  desTime = 20.0;
  desTheta = desDegrees*(PI/180.0);
  
  // Calc the desired steps, first in floats
  nLf = (desRadius - AXLE_LENGTH/2.0)*desTheta/dx;
  nRf = (desRadius + AXLE_LENGTH/2.0)*desTheta/dx;
  nL = (unsigned long) nLf;
  nR = (unsigned long) nRf;
  
  
  robotOmega = desDegrees*(PI/180)/desTime;
  omegaR = (desRadius + AXLE_LENGTH/2.0)*robotOmega/WHEEL_RADIUS;  
  omegaL = (desRadius - AXLE_LENGTH/2.0)*robotOmega/WHEEL_RADIUS;   
  printf("nL = %d, nR = %d\n",nL,nR);  
  printf("omegaL = %f, omegaR = %f\n\n",omegaL,omegaR);
       
  // apply corrections (not for initial run) --------------------
  
  /*
  nL = 3417;
  nR = 4133;
  nLf = (float)nL;
  nRf = (float)nR;
  
  omegaR = omegaL * nRf / nLf;
  
  printf("nL = %d, nR = %d\n",nL,nR);
  printf("omegaL = %f, omegaR = %f\n\n",omegaL,omegaR);
  */
  
  wb_robot_init();
  
  init();
    
  state = STATE_DRIVE;
  
  while (wb_robot_step(TIME_STEP) != -1) {
  
    switch(state) {
    
      case STATE_DRIVE:
        printf("\n");
        
         // Get current wheel angles from the ePuck
        thetaL  = wb_position_sensor_get_value(leftWheelAngle);
        thetaR  = wb_position_sensor_get_value(rightWheelAngle);
        
        countL = ISR_Left();
        countR = ISR_Right();

        dsL = countL*dx;
        dsR = countR*dx;    
        dsC = (dsL + dsR) / 2.0;
        thetaRob = dsC / desRadius;
        
        percY = desRadius*(1 - cos(thetaRob));
        percX = desRadius*sin(thetaRob);
        
        printf("countL = %d, countR = %d thetaRob = %f, dsL = %f, dsR = %f\n",
                countL, countR, thetaRob*180/PI, dsL, dsR);
 
        const double *gpsVals = wb_gps_get_values(gps); 
                    
        writeData4ToOctave(plotFile,percX,percY, gpsVals[0],gpsVals[1]);
             
        // Use counts to halt ------------------------------------------------         
        if( (countL >= nL) && (countR >= nR)) {
          state = STATE_DONE;   
          omegaL = 0.0;
          omegaR = 0.0;
          
          // calculate corrections for next run ----------------------------
          
          // Estimate the arc radius and angle
          double actX = gpsVals[0];
          double actY = gpsVals[1];
          double r = sqrt(actX*actX + (desRadius - actY)*(desRadius - actY));
          double theta = 90 - (180/PI)* atan2((desRadius - actY),actX);
          printf("Estimate of: radius = %f, angle = %f\n",r,theta);         
          
          // Make the correction
          double alpha = desDegrees/theta;
          double beta = desRadius/r;
          
          double nL_new = 0.5*(nRf*alpha*(beta - 1) + nLf*alpha*(beta + 1));
          double nR_new = 0.5*(nRf*alpha*(1 + beta) + nLf*alpha*(beta - 1));
          
          printf("nL_new = %f, nR_new = %f\n",nL_new,nR_new);
  
        }        
                      
        break; 
        
      case STATE_DONE:
        omegaL = 0.0;
        omegaR = 0.0;  
          writeOctaveTail(plotFile);
          fclose(plotFile);
        break;      
    
    }

    wb_motor_set_velocity(left_motor,  omegaL);
    wb_motor_set_velocity(right_motor, omegaR);

  };

  wb_robot_cleanup();

  return 0;
}
